#!/usr/bin/python

import cgi
import datetime
import hashlib
import os
import sys

sys.path.insert(0, os.path.dirname(__file__))
import icommon

output_dir = "/srv/package-import.canonical.com/new/logs/status/"

html_head = '''<html>
<head>
<title>%(title)s</title>
<meta http-equiv="refresh" content="600">
<script src="jquery.min.js"></script>
<script src="jquery.flot.js"></script>
</head>
<body>
<a href="http://package-import.ubuntu.com/failures/.bzr/failures/index.html">Overview</a>
<p>You are looking at information on the Bazaar importer system that serves <a href="https://wiki.ubuntu.com/DistributedDevelopment">Ubuntu Distributed Development</a>. It's hoped that you won't have to care about the existence of this, but things don't always work out that way.</p>
<p>It is expected that uploads appear in the branches in less than 30 minutes, if it takes longer then something is up and this page should tell you why. To dicuss issues or to ask for a package to be looked at please email James Westby.</p>
<p>This page was last updated at <tt>''' + datetime.datetime.utcnow().isoformat() + ''' UTC.</tt></p>
'''
html_tail = '''</body>
</html>
'''


def process_failures():
    db = icommon.StatusDatabase(icommon.sqlite_file)
    return db.summarise_failures()


def package_link(package):
    return '<a href="%s.html#%s">%s</a>' % (cgi.escape(package.name, True),
            package.timestamp, cgi.escape(package.name))


def error_page(reason):
    checksum = hashlib.md5()
    checksum.update(reason)
    return "%s.html" % cgi.escape(checksum.hexdigest(), True)


def error_link(reason):
    return '<a href="%s">Packages with this issue</a>' % error_page(reason)


def list_old_pages():
    old_pages = set()
    for page in os.listdir(output_dir):
        path = os.path.join(output_dir, page)
        if not os.path.isdir(path):
            old_pages.add(path)
    return old_pages


def remove_obsolete_pages(old_pages, new_pages):
    obsolete_pages = old_pages - new_pages
    for path in obsolete_pages:
        if os.path.exists(path) and not os.path.isdir(path):
            os.unlink(path)


def write_individual_pages(reasons, explanations):
    paths = set()
    for reason in reasons:
        for info in reasons[reason][:]:
            assert not "/" in info.name
            path = os.path.join(output_dir, cgi.escape(info.name,
                        True) + ".html")
            f = open(path, "wb")
            try:
                f.write(html_head % {"title":
                        "Failure reason for %s" % cgi.escape(info.name)})
                f.write("<p>Failed at %s</p>\n" % info.timestamp)
                if reason in explanations:
                    f.write("<p>%s</p>\n" % cgi.escape(explanations[reason]))
                f.write("<p><pre>\n")
                f.write(cgi.escape(info.raw_reason))
                f.write("\n</pre></p>\n")
                f.write("<div>%s</div>" % error_link(reason))
                f.write(html_tail)
            finally:
                f.close()
            paths.add(path)
        path = os.path.join(output_dir, error_page(reason))
        f = open(path, "wb")
        try:
            f.write(html_head % {"title": "Failures of type"})
            f.write("<ul>\n")
            for package in reasons[reason][:]:
                f.write("<li>%s</li>\n" % package_link(package))
            f.write("</ul>\n")
            f.write(html_tail)
        finally:
            f.close()
        paths.add(path)
    return paths


def find_info_where(package_info, clause):
    ret = []
    for info in package_info:
        if clause(info):
            ret.append(info)
    return ret


def write_stats(f, package_info):
    f.write("<ul>\n")
    running = find_info_where(package_info, lambda x: x.running)
    f.write("<li>%d currently running\n<ul>\n" % len(running))
    running.sort(key=lambda x: x.timestamp)
    for package in running:
        f.write("<li>%s (since %s)</li>" % (cgi.escape(package.name),
                package.timestamp))
    f.write("</ul>\n</li>\n")
    outstanding = find_info_where(package_info, lambda x: x.queued)
    f.write("<li>%d outstanding jobs\n<ul>\n<li>" % len(outstanding))
    outstanding.sort(key=lambda x: x.name)
    for package in outstanding:
        f.write(cgi.escape(package.name) + "&nbsp;")
    f.write("</li>\n</ul>\n</li>\n")
    f.write("<li>%d failures\n</li>"
            % (len(package_info) - len(outstanding) - len(running)))
    f.write("</ul>\n")


def write_analyis_partial(f, keys, reasons, explanations, clause, text):
    for reason in keys:
        packages = reasons[reason]
        subset = find_info_where(packages, clause)
        if len(subset) > 0:
            f.write("<li>%d packages %s with key %s\n<ul>\n"
                    % (len(subset), text, cgi.escape(reason)))
            if reason in explanations:
                f.write('<li><pre>%s</pre></li>\n'
                        % cgi.escape(explanations[reason]))
            f.write('<li>')
            subset.sort(key=lambda x: x.name)
            for package in subset:
                f.write(package_link(package) + "&nbsp;")
            f.write("</li>\n</ul>\n</li>\n")


def write_analysis(f, reasons, explanations):
    keys = reasons.keys()
    def compare(a, b):
        return len(reasons[b]) - len(reasons[a])
    keys.sort(cmp=compare)
    write_analyis_partial(f, keys, reasons, explanations,
            lambda x: x.auto_retry_masked, "failed to many times to retry")
    write_analyis_partial(f, keys, reasons, explanations,
            lambda x: not x.auto_retry, "failed")
    write_analyis_partial(f, keys, reasons, explanations,
            lambda x: x.auto_retry and not x.auto_retry_masked,
            "spuriously failed")

def write_latest(f, package_info):
    failed = find_info_where(package_info, lambda x: x.failure_count > 0)
    failed.sort(key=lambda x: x.timestamp, reverse=True)
    i = 0
    for package in failed:
        i += 1
        if i > 50:
            return
        f.write("<li>%s\n<ul>\n" % package_link(package))
        f.write("<li>failed at %s</li>\n" % package.timestamp)
        f.write("<li>with key %s</li>\n" % cgi.escape(package.signature))
        if package.auto_retry_masked:
            f.write("<li>and failed too many times to retry</li>\n")
        elif package.auto_retry:
            f.write("<li>and will be retried automatically around %s</li>\n" % package.auto_retry_time)
        f.write("</ul>\n</li>\n")

def write_overview_page(reasons, package_info, explanations):
    paths = set()
    path = os.path.join(output_dir, "index.html")
    f = open(path, "wb")
    try:
        f.write(html_head % {"title": "bzr import failures"})
        f.write("""<p><h3>Contents of this page</h3>
<ul>
<li><a href="#stats">Summary</a></li>
<li><a href="#latest">Latest failures</a></li>
<li><a href="#analysis">All failures</a></li>
</ul>
</p>
""")
        f.write('<div id="stats">\n')
        write_stats(f, package_info)
        f.write('</div>\n<div id="latest">\n<p>Latest 50 Failures:\n<ul>\n')
        write_latest(f, package_info)
        f.write('</ul>\n</p>\n</div>\n')
        f.write('</div>\n<div id="analysis">\n<ul>\n')
        write_analysis(f, reasons, explanations)
        f.write('</ul>\n</div>\n')
        f.write(html_tail)
    finally:
        f.close()
    paths.add(path)
    return paths


def get_main_packages():
    db = icommon.PackageDatabase(icommon.sqlite_package_file)
    return db.list_packages_in_main()


def filter_packages(filter_set, reasons, package_info):
    filtered_package_info = find_info_where(package_info, lambda x: x.name in filter_set)
    filtered_reasons = {}
    for sig in reasons:
        for info in reasons[sig]:
            if info in filtered_package_info:
                filtered_reasons.setdefault(sig, [])
                filtered_reasons[sig].append(info)
    return filtered_reasons, filtered_package_info


def write_main_page(reasons, package_info, explanations):
    main_packages = get_main_packages()
    reasons, package_info = filter_packages(main_packages, reasons, package_info)
    paths = set()
    path = os.path.join(output_dir, "main.html")
    f = open(path, "wb")
    try:
        f.write(html_head % {"title": "bzr import failures for main"})
        f.write("""<p><h3>Contents of this page</h3>
<ul>
<li><a href="#stats">Summary</a></li>
<li><a href="#latest">Latest failures</a></li>
<li><a href="#analysis">All failures</a></li>
</ul>
</p>
""")
        f.write('<div id="stats">\n')
        write_stats(f, package_info)
        f.write('</div>\n<div id="latest">\n<p>Latest 50 Failures:\n<ul>\n')
        write_latest(f, package_info)
        f.write('</ul>\n</p>\n</div>\n')
        f.write('</div>\n<div id="analysis">\n<ul>\n')
        write_analysis(f, reasons, explanations)
        f.write('</ul>\n</div>\n')
        f.write(html_tail)
    finally:
        f.close()
    paths.add(path)
    return paths


def get_info():
    new_output = None
    reasons, package_info = process_failures()
    explanations = icommon.load_explanations()
    old_pages = list_old_pages()
    new_pages = write_individual_pages(reasons, explanations)
    new_pages |= write_overview_page(reasons, package_info, explanations)
    new_pages |= write_main_page(reasons, package_info, explanations)
    remove_obsolete_pages(old_pages, new_pages)


def main():
    lock = icommon.lock_categorise_failures()
    if lock is None:
        print "Another instance of categorise_failures is already running."
        sys.exit(0)
    try:
        get_info()
    finally:
        lock.close()


if __name__ == '__main__':
    main()
