#!/usr/bin/python

import os
import sys
import unittest

from bzrlib.tests import TestCaseWithTransport
from bzrlib import transform

from debian_bundle import changelog

sys.path.insert(0, os.path.dirname(__file__))
import icommon
import import_package


class PackageToImportTests(unittest.TestCase):

    def test_same_as_same(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="main",
                on_lp=True, url=None)
        b = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="contrib")
        c = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", on_lp=False)
        d = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", url="a")
        self.assertTrue(a.same_as(b))
        self.assertTrue(a.same_as(c))
        self.assertTrue(a.same_as(d))

    def test_same_as_different(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="main",
                on_lp=True, url=None)
        b = icommon.PackageToImport("b", changelog.Version("1"),
                "debian", "sid", "release")
        c = icommon.PackageToImport("a", changelog.Version("1.1"),
                "debian", "sid", "release")
        d = icommon.PackageToImport("a", changelog.Version("1"),
                "ubuntu", "sid", "release")
        e = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "squeeze", "release")
        f = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "security")
        self.assertFalse(a.same_as(b))
        self.assertFalse(a.same_as(c))
        self.assertFalse(a.same_as(d))
        self.assertFalse(a.same_as(e))
        self.assertFalse(a.same_as(f))

    def test_get_url(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", url="url")
        self.assertEqual("url", a.get_url())
        b = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="non-free", on_lp=False)
        self.assertEqual(icommon.urllib_debian_base_url
                + "pool/non-free/a/a/a_1.dsc", b.get_url())
        c = icommon.PackageToImport("libfoo", changelog.Version("1:1.1-1"),
                "debian", "squeeze", "release", component="contrib", on_lp=False)
        self.assertEqual(icommon.urllib_debian_base_url
                + "pool/contrib/libf/libfoo/libfoo_1.1-1.dsc", c.get_url())
        c = icommon.PackageToImport("libfoo", changelog.Version("1:1.1-1"),
                "debian", "squeeze", "release", component="contrib",
                on_lp=True)
        self.assertEqual(icommon.urllib_launchpad_base_url
                + "debian/squeeze/+source/libfoo/1:1.1-1/+files/libfoo_1.1-1.dsc",
                c.get_url())
        d = icommon.PackageToImport("libfoo", changelog.Version("1:1.1-1"),
                "ubuntu", "lucid", "security")
        self.assertEqual(icommon.urllib_launchpad_base_url
                + "ubuntu/lucid/+source/libfoo/1:1.1-1/+files/libfoo_1.1-1.dsc",
                d.get_url())

    def test_str(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="main",
                on_lp=True, url=None)
        self.assertEqual("PackageToImport(a, 1, debian, sid, release, main)",
                str(a))


class ImportListTests(unittest.TestCase):

    def make_simple(self):
        self.pkg1 = icommon.PackageToImport("pkg", "0.1-1", "ubuntu",
                "jaunty", "Release")
        self.pkg2 = icommon.PackageToImport("pkg", "1.0-1", "ubuntu",
                "karmic", "Release")
        list = icommon.ImportList()
        list.add_if_needed(self.pkg1)
        list.add_if_needed(self.pkg2)
        return list

    def test_str(self):
        list = self.make_simple()
        self.assertEqual("[PackageToImport(pkg, 0.1-1, ubuntu, "
                "jaunty, Release, ), PackageToImport(pkg, 1.0-1, "
                "ubuntu, karmic, Release, )]", str(list))

    def test_reverse(self):
        list = self.make_simple()
        self.assertEqual([self.pkg1, self.pkg2], list.plist)
        list.reverse()
        self.assertEqual([self.pkg2, self.pkg1], list.plist)

    def test_sort(self):
        list = self.make_simple()
        self.assertEqual([self.pkg1, self.pkg2], list.plist)
        list.sort()
        self.assertEqual([self.pkg2, self.pkg1], list.plist)

    def test_add_if_needed(self):
        list = self.make_simple()
        self.assertEqual([self.pkg1, self.pkg2], list.plist)
        list.add_if_needed(self.pkg1)
        self.assertEqual([self.pkg1, self.pkg2], list.plist)


class RevidDatabaseTests(TestCaseWithTransport):

    def setUp(self):
        super(RevidDatabaseTests, self).setUp()
        self.cleanup = 0
        self.push = 0
        self.db = self.get_db()
        self.version = changelog.Version("1.0-1")
        self.version2 = changelog.Version("1.0-2")
        self.version3 = changelog.Version("1.0-3")
        self.suite = "karmic"
        self.tree = self.make_branch_and_tree(".")
        self.revid = self.tree.commit("one")
        self.revid2 = self.tree.commit("two")
        self.revid3 = self.tree.commit("three")

    def get_db(self):
        return icommon.RevidDatabase(":memory:", "pkg")

    def check_rows(self, memory, working, saved):
        self.assertEqual(memory, len(self.db._memory_rows()))
        self.assertEqual(working, len(self.db._working_rows()))
        self.assertEqual(saved, len(self.db._saved_rows()))

    def test_mark(self):
        self.assertFalse(self.db.is_marked(self.version, self.suite))
        self.check_rows(0, 0, 0)
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.check_rows(1, 0, 0)
        self.assertTrue(self.db.is_marked(self.version, self.suite))

    def test_mark_again(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.assertRaises(AssertionError, self.db.commit_outstanding)
        self.assertRaises(AssertionError, self.db.is_marked, self.version,
                self.suite)

    def test_check(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.assertTrue(self.db.check(self.version, self.revid, self.suite,
                    self.tree.branch))
        self.assertFalse(self.db.check(changelog.Version(
                        str(self.version)+"1"),
                    self.revid, self.suite, self.tree.branch))
        self.assertFalse(self.db.check(self.version, self.revid,
                    self.suite+"a", self.tree.branch))

    def test_commit(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.check_rows(0, 0, 1)
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.assertRaises(AssertionError, self.db.commit_outstanding)
        self.check_rows(1, 0, 1)

    def do_cleanup(self):
        self.cleanup += 1

    def do_push(self):
        self.push += 1

    def test_cleanup_nothing_to_do(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit()
        self.db.cleanup_last_run(self.do_cleanup, self.do_push)
        self.assertEqual(0, self.push)
        self.assertEqual(1, self.cleanup)

    def test_cleanup(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.db.mark(self.version2, self.revid2, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.check_rows(0, 1, 1)
        self.db.cleanup_last_run(self.do_cleanup, self.do_push)
        self.assertEqual(1, self.push)
        self.assertEqual(1, self.cleanup)
        self.check_rows(0, 0, 2)

    def test_discard(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.mark(self.version2, self.revid2, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.db.mark(self.version3, self.revid3, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.check_rows(0, 1, 2)
        self.db.discard_last_run()
        self.check_rows(0, 0, 2)


class FindEarliestMergeTests(TestCaseWithTransport):

    def test_tip_revision(self):
        t = self.make_branch_and_tree(".")
        revid = t.commit("one")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, revid))

    def test_a_parent(self):
        t = self.make_branch_and_tree(".")
        revid = t.commit("one")
        t.commit("two")
        t.commit("three")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, revid))

    def test_merged(self):
        t = self.make_branch_and_tree(".")
        t.commit("one")
        t2 = t.bzrdir.sprout("merged").open_workingtree()
        merged_revid = t2.commit("one and a half")
        t.merge_from_branch(t2.branch)
        revid = t.commit("two")
        t.commit("three")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, merged_revid))


class TestCheckSame(TestCaseWithTransport):

    def assertTreesEquivalent(self, repo, rev1, rev2):
        rtree1 = repo.revision_tree(rev1)
        rtree2 = repo.revision_tree(rev2)
        self.assertTrue(
            import_package._are_trees_equivalent(rtree1, rtree2),
            'Trees are not actually equivalent')

    def assertTreesNotEquivalent(self, repo, rev1, rev2):
        rtree1 = repo.revision_tree(rev1)
        rtree2 = repo.revision_tree(rev2)
        self.assertFalse(
            import_package._are_trees_equivalent(rtree1, rtree2),
            'Trees are accidentally equivalent')

    def test_identical_different_revs_same_fids(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/file', 'content\n')])
        t1.add(['file'], ['file-id'])
        revid = t1.commit('first')
        t2 = t1.bzrdir.sprout('tree2').open_workingtree()
        # Both trees independently get identical content
        self.build_tree_contents([('tree1/file', 'new content\n'),
                                  ('tree2/file', 'new content\n')])
        t1_rev = t1.commit('new content')
        t2_rev = t2.commit('new content')
        t1.branch.repository.fetch(t2.branch.repository, t2_rev)

        self.assertTreesEquivalent(repo, t1_rev, t2_rev)

    def test_exec_patch_change_is_ignored(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/patches/',),
                                  ('tree1/debian/patches/my.patch', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/patches', 'debian/patches/my.patch'])
        rev1 = t1.commit('create the new tree')
        tt = transform.TreeTransform(t1)
        tt.set_executability(True,
            tt.trans_id_file_id(t1.path2id('debian/patches/my.patch')))
        tt.apply()
        rev2 = t1.commit('set executable bit')
        self.assertTreesEquivalent(repo, rev1, rev2)
        self.build_tree_contents([
            ('tree1/debian/patches/my.patch', 'new content\n'),])
        rev3 = t1.commit('change content')
        self.assertTreesNotEquivalent(repo, rev1, rev3)
        self.assertTreesNotEquivalent(repo, rev2, rev3)

    def test_deleting_content_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/file', 'content\n'),
                                  ('tree1/debian/dir/file2', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/file', 'debian/dir/file2'])
        rev1 = t1.commit('initial')
        os.remove('tree1/debian/dir/file2')
        rev2 = t1.commit('del file2')
        self.assertTreesNotEquivalent(repo, rev1, rev2)

    def test_deleting_last_content_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/file', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/file'])
        rev1 = t1.commit('initial')
        os.remove('tree1/debian/dir/file')
        rev2 = t1.commit('del file2')
        self.assertTreesNotEquivalent(repo, rev1, rev2)

    def test_deleting_cascaded_dirs_is_not_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/subdir1/',),
                                  ('tree1/debian/dir/subdir2/',),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/subdir1',
                'debian/dir/subdir2'])
        rev1 = t1.commit('initial')
        os.rmdir('tree1/debian/dir/subdir1')
        os.rmdir('tree1/debian/dir/subdir2')
        os.rmdir('tree1/debian/dir')
        rev2 = t1.commit('del subtree')
        self.assertTreesEquivalent(repo, rev1, rev2)
        # The reverse change is *not* equivalent
        self.assertTreesNotEquivalent(repo, rev2, rev1)

    def test_deleting_non_debian_empty_dir_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/subdir1/',),
                                  ])
        t1.add(['debian', 'debian/dir', 'subdir1'])
        rev1 = t1.commit('initial')
        os.rmdir('tree1/subdir1')
        rev2 = t1.commit('del subbdir')
        self.assertTreesNotEquivalent(repo, rev1, rev2)


if __name__ == '__main__':
    unittest.main()

